# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.__internal__ namespace.
"""

import sys as _sys

from keras.api._v2.keras.__internal__ import backend
from keras.api._v2.keras.__internal__ import layers
from keras.api._v2.keras.__internal__ import losses
from keras.api._v2.keras.__internal__ import models
from keras.api._v2.keras.__internal__ import utils
from keras.engine.base_layer import _apply_name_scope_on_model_declaration as apply_name_scope_on_model_declaration