# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Keras mixed precision API.

See [the mixed precision guide](
  https://www.tensorflow.org/guide/keras/mixed_precision) to learn how to
use the API.

"""

import sys as _sys

from keras.mixed_precision.loss_scale_optimizer import BaseLossScaleOptimizer as LossScaleOptimizer
from keras.mixed_precision.policy import Policy
from keras.mixed_precision.policy import global_policy
from keras.mixed_precision.policy import set_global_policy